from hybrid_pool_executor.base import Future
from hybrid_pool_executor.executor import HybridPoolExecutor
from hybrid_pool_executor.spec import spec_factory as SpecFactory

__version__ = "0.0.4"
__author__ = "Leavers"
__contact__ = "leavers930@gmail.com"

__all__ = (
    "Future",
    "HybridPoolExecutor",
    "SpecFactory",
)
